#include "otsdaq-core/ConfigurationInterface/ConfigurationManager.h"
#include "otsdaq-core/Macros/TablePluginMacros.h"
#include "otsdaq-core/TablePlugins/DesktopIconTable.h"
#include "otsdaq-core/WebUsersUtilities/WebUsers.h"

#include <stdio.h>
#include <fstream>  // std::fstream
#include <iostream>
using namespace ots;

#define DESKTOP_ICONS_FILE \
	std::string(__ENV__("SERVICE_DATA_PATH")) + "/OtsWizardData/iconList.dat"

// DesktopIconTable Column names
#define COL_NAME "IconName"
#define COL_STATUS TableViewColumnInfo::COL_NAME_STATUS
#define COL_CAPTION "Caption"
#define COL_ALTERNATE_TEXT "AlternateText"
#define COL_FORCE_ONLY_ONE_INSTANCE "ForceOnlyOneInstance"
#define COL_REQUIRED_PERMISSION_LEVEL "RequiredPermissionLevel"
#define COL_IMAGE_URL "ImageURL"
#define COL_WINDOW_CONTENT_URL "WindowContentURL"
#define COL_APP_LINK "LinkToApplicationTable"
#define COL_PARAMETER_LINK "LinkToParameterTable"
#define COL_PARAMETER_KEY "windowParameterKey"
#define COL_PARAMETER_VALUE "windowParameterValue"
#define COL_FOLDER_PATH "FolderPath"

// XDAQ App Column names
#define COL_APP_ID "Id"

//==============================================================================
DesktopIconTable::DesktopIconTable(void) : TableBase("DesktopIconTable")
{
	// Icon list no longer passes through file! so delete it from user's $USER_DATA
	std::system(("rm -rf " + (std::string)DESKTOP_ICONS_FILE).c_str());

	//////////////////////////////////////////////////////////////////////
	// WARNING: the names used in C++ MUST match the Table INFO  //
	//////////////////////////////////////////////////////////////////////
}

//==============================================================================
DesktopIconTable::~DesktopIconTable(void) {}

//==============================================================================
void DesktopIconTable::init(ConfigurationManager* configManager)
{
	//	__COUT__ << "*&*&*&*&*&*&*&*&*&*&*&*&*&*&*&*&*&*&*&*&*&*" << std::endl;
	//	__COUT__ << configManager->__SELF_NODE__ << std::endl;

	unsigned int intVal;

	auto childrenMap = configManager->__SELF_NODE__.getChildren();

	activeDesktopIcons_.clear();

	DesktopIconTable::DesktopIcon* icon;
	bool                           addedAppId;
	bool                           numeric;
	unsigned int                   i;
	for(auto& child : childrenMap)
	{
		if(!child.second.getNode(COL_STATUS).getValue<bool>())
			continue;

		activeDesktopIcons_.push_back(DesktopIconTable::DesktopIcon());
		icon = &(activeDesktopIcons_.back());

		icon->caption_ = child.second.getNode(COL_CAPTION).getValue<std::string>();
		icon->alternateText_ =
		    child.second.getNode(COL_ALTERNATE_TEXT).getValue<std::string>();
		icon->enforceOneWindowInstance_ =
		    child.second.getNode(COL_FORCE_ONLY_ONE_INSTANCE).getValue<bool>();
		icon->permissionThresholdString_ =
		    child.second.getNode(COL_REQUIRED_PERMISSION_LEVEL).getValue<std::string>();
		icon->imageURL_ = child.second.getNode(COL_IMAGE_URL).getValue<std::string>();
		icon->windowContentURL_ =
		    child.second.getNode(COL_WINDOW_CONTENT_URL).getValue<std::string>();
		icon->folderPath_ = child.second.getNode(COL_FOLDER_PATH).getValue<std::string>();

		if(icon->folderPath_ == TableViewColumnInfo::DATATYPE_STRING_DEFAULT)
			icon->folderPath_ = "";  // convert DEFAULT to empty string

		if(icon->permissionThresholdString_ ==
		   TableViewColumnInfo::DATATYPE_STRING_DEFAULT)
			icon->permissionThresholdString_ =
			    "1";  // convert DEFAULT to standard user allow

		numeric = true;
		for(i = 0; i < icon->permissionThresholdString_.size(); ++i)
			if(!(icon->permissionThresholdString_[i] >= '0' &&
			     icon->permissionThresholdString_[i] <= '9'))
			{
				numeric = false;
				break;
			}
		// for backwards compatibility, if permissions threshold is a single number
		//	assume it is the threshold intended for the WebUsers::DEFAULT_USER_GROUP group
		if(numeric)
			icon->permissionThresholdString_ =
			    WebUsers::DEFAULT_USER_GROUP + ":" + icon->permissionThresholdString_;

		// remove all commas from member strings because desktop icons are served to
		// client in comma-separated string
		icon->caption_ = removeCommas(
		    icon->caption_, false /*andHexReplace*/, true /*andHTMLReplace*/);
		icon->alternateText_ = removeCommas(
		    icon->alternateText_, false /*andHexReplace*/, true /*andHTMLReplace*/);
		icon->imageURL_ = removeCommas(icon->imageURL_, true /*andHexReplace*/);
		icon->windowContentURL_ =
		    removeCommas(icon->windowContentURL_, true /*andHexReplace*/);
		icon->folderPath_ = removeCommas(
		    icon->folderPath_, false /*andHexReplace*/, true /*andHTMLReplace*/);

		// add URN/LID to windowContentURL_, if link is given
		addedAppId = false;
		if(!child.second.getNode(COL_APP_LINK).isDisconnected())
		{
			// if last character is not '='
			//	then assume need to add "?urn="
			if(icon->windowContentURL_[icon->windowContentURL_.size() - 1] != '=')
				icon->windowContentURL_ += "?urn=";

			//__COUT__ << "Following Application link." << std::endl;
			child.second.getNode(COL_APP_LINK).getNode(COL_APP_ID).getValue(intVal);
			icon->windowContentURL_ += std::to_string(intVal);

			//__COUT__ << "URN/LID=" << intVal << std::endl;
			addedAppId = true;
		}

		// add parameters if link is given
		if(!child.second.getNode(COL_PARAMETER_LINK).isDisconnected())
		{
			// if there is no '?' found
			//	then assume need to add "?"
			if(icon->windowContentURL_.find('?') == std::string::npos)
				icon->windowContentURL_ += '?';
			else if(addedAppId ||
			        icon->windowContentURL_[icon->windowContentURL_.size() - 1] !=
			            '?')  // if not first parameter, add &
				icon->windowContentURL_ += '&';

			// now add each paramter separated by &
			auto paramGroupMap = child.second.getNode(COL_PARAMETER_LINK).getChildren();
			bool notFirst      = false;
			for(const auto param : paramGroupMap)
			{
				if(notFirst)
					icon->windowContentURL_ += '&';
				else
					notFirst = true;
				icon->windowContentURL_ +=
				    ConfigurationManager::encodeURIComponent(
				        param.second.getNode(COL_PARAMETER_KEY).getValue<std::string>()) +
				    "=" +
				    ConfigurationManager::encodeURIComponent(
				        param.second.getNode(COL_PARAMETER_VALUE)
				            .getValue<std::string>());
			}
		}
	}  // end main icon extraction loop

	//
	//	//generate icons file
	//	std::fstream fs;
	//	fs.open(DESKTOP_ICONS_FILE, std::fstream::out | std::fstream::trunc);
	//	if(fs.fail())
	//	{
	//		__SS__ << "Failed to open Desktop Icons run file: " << DESKTOP_ICONS_FILE <<
	// std::endl;
	//		__SS_THROW__;
	//	}
	//
	//	for(auto &child:childrenMap)
	//	{
	//		child.second.getNode(COL_STATUS	).getValue(status);
	//		if(!status) continue;
	//
	//		if(first) first = false;
	//		else fs << ",";
	//
	//		child.second.getNode(COL_CAPTION	).getValue(val);
	//		fs << removeCommas(val, false, true);
	//		//__COUT__ << "Icon caption: " << val << std::endl;
	//
	//		fs << ",";
	//		child.second.getNode(COL_ALTERNATE_TEXT	).getValue(val);
	//		fs << removeCommas(val, false, true);
	//
	//		fs << ",";
	//		child.second.getNode(COL_FORCE_ONLY_ONE_INSTANCE	).getValue(status);
	//		fs << (status?"1":"0");
	//
	//		fs << ",";
	//		child.second.getNode(COL_REQUIRED_PERMISSION_LEVEL	).getValue(val);
	//		fs << removeCommas(val);
	//
	//		fs << ",";
	//		child.second.getNode(COL_IMAGE_URL	).getValue(val);
	//		fs << removeCommas(val,true);
	//
	//		fs << ",";
	//		child.second.getNode(COL_WINDOW_CONTENT_URL	).getValue(val);
	//		val = removeCommas(val,true);
	//		fs << val;
	//
	//		bool addedAppId = false;
	//		//add URN/LID if link is given
	//		if(!child.second.getNode(COL_APP_LINK	).isDisconnected())
	//		{
	//			//if last character is not '='
	//			//	then assume need to add "?urn="
	//			if(val[val.size()-1] != '=')
	//				fs << "?urn=";
	//
	//			//__COUT__ << "Following Application link." << std::endl;
	//			child.second.getNode(COL_APP_LINK	).getNode(COL_APP_ID
	//).getValue(intVal);
	//
	//			//__COUT__ << "URN/LID=" << intVal << std::endl;
	//			fs << intVal; //append number
	//			addedAppId = true;
	//		}
	//
	//		//add parameters if link is given
	//		if(!child.second.getNode(COL_PARAMETER_LINK	).isDisconnected())
	//		{
	//			//if there is no '?' found
	//			//	then assume need to add "?"
	//			if(val.find('?') == std::string::npos)
	//				fs << '?';
	//			else if(addedAppId ||
	//					val[val.size()-1] != '?') //if not first parameter, add &
	//				fs << '&';
	//
	//			//now add each paramter separated by &
	//			auto paramGroupMap = child.second.getNode(COL_PARAMETER_LINK
	//).getChildren(); 			bool notFirst = false; 			for(const auto
	// param:paramGroupMap)
	//			{
	//				if(notFirst)
	//					fs << '&';
	//				else
	//					notFirst = true;
	//				fs << ConfigurationManager::encodeURIComponent(
	//						param.second.getNode(COL_PARAMETER_KEY).getValue<std::string>())
	//<<
	//"="
	//<< 								ConfigurationManager::encodeURIComponent(
	//										param.second.getNode(COL_PARAMETER_VALUE).getValue<std::string>());
	//			}
	//		}
	//
	//		fs << ",";
	//		child.second.getNode(COL_FOLDER_PATH	).getValue(val);
	//		if(val == TableViewColumnInfo::DATATYPE_STRING_DEFAULT) val = "";
	// 		fs << removeCommas(val,true);
	//	}
	//
	//	//close icons file
	//	fs.close();
}

std::string DesktopIconTable::removeCommas(const std::string& str,
                                           bool               andHexReplace,
                                           bool               andHTMLReplace)
{
	std::string retStr = "";
	retStr.reserve(str.length());

	for(unsigned int i = 0; i < str.length(); ++i)
		if(str[i] != ',')
			retStr += str[i];
		else if(andHexReplace)
			retStr += "%2C";
		else if(andHTMLReplace)
			retStr += "&#44;";

	return retStr;
}

DEFINE_OTS_TABLE(DesktopIconTable)
